import logging
from uuid import UUID

from django.conf import settings
from django.core.mail import send_mail
from django.db import transaction
from django.db.models import CharField, Count, F, Max, Q, Value
from django.db.models.functions import Cast, Concat, Greatest, NullIf
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.utils.functional import cached_property
from drf_spectacular.utils import (
    OpenApiExample,
    OpenApiParameter,
    PolymorphicProxySerializer,
    extend_schema,
    extend_schema_view,
)
from rest_framework import permissions, status
from rest_framework.exceptions import NotFound, PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    GenericAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveDestroyAPIView,
    RetrieveUpdateAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.response import Response
from rest_framework.views import APIView

from arkindex.dataimport.models import (
    DataFile,
    DataImport,
    DataImportMode,
    GitRefType,
    Repository,
    RepositoryType,
    Revision,
    Worker,
    WorkerActivity,
    WorkerActivityState,
    WorkerRun,
    WorkerVersion,
)
from arkindex.dataimport.serializers.files import DataFileCreateSerializer, DataFileSerializer
from arkindex.dataimport.serializers.git import ExternalRepositorySerializer, RevisionSerializer
from arkindex.dataimport.serializers.imports import (
    CreateImportTranskribusErrorResponseSerializer,
    DataImportFromFilesSerializer,
    DataImportLightSerializer,
    DataImportSerializer,
    ElementsWorkflowSerializer,
    ImportTranskribusSerializer,
    ProcessElementLightSerializer,
    ProcessElementSerializer,
    StartProcessSerializer,
    WorkerRunEditSerializer,
    WorkerRunSerializer,
)
from arkindex.dataimport.serializers.workers import (
    RepositorySerializer,
    WorkerActivitySerializer,
    WorkerSerializer,
    WorkerVersionEditSerializer,
    WorkerVersionSerializer,
)
from arkindex.documents.models import Corpus, Element
from arkindex.project.fields import ArrayRemove
from arkindex.project.mixins import (
    CorpusACLMixin,
    CustomPaginationViewMixin,
    DeprecatedMixin,
    ProcessACLMixin,
    RepositoryACLMixin,
    SelectionMixin,
    WorkerACLMixin,
)
from arkindex.project.permissions import IsInternalUser, IsVerified, IsVerifiedOrReadOnly
from arkindex.project.tools import RTrimChr
from arkindex.users.models import OAuthCredentials, Role
from arkindex.users.utils import get_max_level
from ponos.models import STATES_ORDERING, State

logger = logging.getLogger(__name__)


@extend_schema_view(
    get=extend_schema(
        operation_id='ListDataImports',
        tags=['imports'],
        parameters=[
            OpenApiParameter(
                'corpus',
                type=UUID,
                description="Filter imports by corpus ID",
                required=False,
            ),
            OpenApiParameter(
                'created',
                type=bool,
                description='Restrict to or exclude imports you created',
                required=False,
            ),
            OpenApiParameter(
                'mode',
                enum=[mode.value for mode in DataImportMode],
                description="Filter imports by mode",
                required=False,
            ),
            OpenApiParameter(
                'state',
                enum=[state.value for state in State],
                description='Filter imports by workflow state',
                required=False,
            ),
            OpenApiParameter(
                'id',
                description='Filter imports by beginning of UUID',
                required=False,
            ),
            OpenApiParameter(
                'with_workflow',
                type=bool,
                description='Restrict to or exclude imports with workflows',
                required=False,
            )
        ]
    )
)
class DataImportsList(ProcessACLMixin, ListAPIView):
    """
    List all visible data imports.
    """
    permission_classes = (IsVerified, )
    serializer_class = DataImportLightSerializer

    def get_queryset(self):
        filters = Q()

        if 'with_workflow' in self.request.query_params:
            with_workflow = self.request.query_params['with_workflow']
            filters &= Q(workflow__isnull=bool(with_workflow and with_workflow.lower() in ('false', '0')))

        if 'corpus' in self.request.query_params:
            corpus_id = self.request.query_params['corpus']
            try:
                corpus_id = UUID(corpus_id)
            except (AttributeError, ValueError):
                raise ValidationError({'corpus': ["'{}' is not a valid UUID".format(corpus_id)]})
            # No supplementary validation is required on the corpus ID filter
            filters &= Q(corpus=corpus_id)

        if 'mode' in self.request.query_params:
            try:
                filters &= Q(mode=DataImportMode(self.request.query_params['mode']))
            except ValueError:
                raise ValidationError({
                    'mode': ["Mode '{}' does not exist".format(self.request.query_params['mode'])]
                })

        if 'created' in self.request.query_params:
            created = self.request.query_params['created']
            creator_filter = Q(creator_id=self.request.user.id)
            if created.lower() in ('false', '0'):
                creator_filter = ~creator_filter
            filters &= creator_filter

        if 'id' in self.request.query_params:
            filters &= Q(id__startswith=self.request.query_params['id'])

        qs = self.readable_processes \
            .filter(filters) \
            .prefetch_related('workflow__tasks')
        # Order workflow by date of last updated task in workflow
        qs = qs.annotate(
            last_date=Greatest(Max('workflow__tasks__updated'), 'updated'),
            last_run=Max('workflow__tasks__run')
        )

        state_value = self.request.query_params.get('state')
        if state_value :
            try:
                state = State(self.request.query_params['state'])
            except ValueError:
                raise ValidationError({'state': [f"State '{state_value}' does not exist"]})
            # Filter out processes which have a task with an incompatible state on their last run
            excluding_states = STATES_ORDERING[:STATES_ORDERING.index(state)]
            excluded_imports = qs.filter(
                Q(workflow__tasks__run=F('last_run')),
                Q(workflow__tasks__state__in=excluding_states)
            )
            # Keep non excluded processes matching the state on their last run tasks
            state_query = Q(workflow__tasks__run=F('last_run')) & Q(workflow__tasks__state=state)
            if state == State.Unscheduled:
                # Handle a workflow with no task as unscheduled
                state_query |= Q(workflow__tasks__isnull=True)
            qs = qs.filter(state_query).exclude(id__in=excluded_imports.values('id')).distinct()

        return qs.order_by('-last_date')


@extend_schema(tags=['imports'])
@extend_schema_view(
    get=extend_schema(description='Retrieve details of a process.\n\nRequires a **guest** access.'),
    patch=extend_schema(description='Partially update a process.\n\nRequires an **admin** access.'),
    put=extend_schema(description='Update dependencies of a process.\n\nRequires an **admin** access.'),
    delete=extend_schema(
        description=(
            'Delete a process.\n\n'
            'Cannot be used on currently running data imports. '
            'Requires an **admin** access.'
        )
    ),
)
class DataImportDetails(ProcessACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve or edit a process
    """
    permission_classes = (IsVerified, )
    serializer_class = DataImportSerializer

    def get_queryset(self):
        return DataImport.objects \
            .select_related('corpus', 'workflow') \
            .prefetch_related('workflow__tasks') \
            .annotate(last_run=Max('workflow__tasks__run'))

    def get_object(self):
        if not hasattr(self, '_dataimport'):
            self._dataimport = super().get_object()
        return self._dataimport

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request and 'pk' in self.kwargs:
            context['dataimport'] = self.get_object()
        return context

    def check_object_permissions(self, request, process):
        super().check_object_permissions(request, process)

        access_level = self.process_access_level(process)
        required_access = Role.Admin.value
        if request.method in permissions.SAFE_METHODS:
            # A guest role allow to retrieve information on a process
            required_access = Role.Guest.value

        if not access_level:
            raise NotFound
        if access_level < required_access:
            raise PermissionDenied(detail='You do not have a sufficient access level to this process.')

        if request.method == 'DELETE' and process.state == State.Running:
            raise ValidationError({'__all__': ['Cannot delete a workflow while it is running']})


class DataImportRetry(ProcessACLMixin, GenericAPIView):
    """
    Retry a data import. Can only be used on imports with Error, Failed, Stopped or Completed states.\n\n
    Requires an **admin** access to the process.
    """
    permission_classes = (IsVerified, )
    serializer_class = DataImportSerializer
    queryset = DataImport.objects.all()

    def check_object_permissions(self, request, process):
        super().check_object_permissions(request, process)

        access_level = self.process_access_level(process)
        if not access_level:
            raise NotFound
        if access_level < Role.Admin.value:
            raise PermissionDenied(detail='You do not have an admin access to this process.')

        if process.workflow is not None and process.state in (State.Unscheduled, State.Pending):
            raise ValidationError({'__all__': ['This workflow is already pending']})
        elif process.state == State.Running:
            raise ValidationError({'__all__': ['This workflow is already running']})

    @extend_schema(
        operation_id='RetryDataImport',
        tags=['imports'],
        request=None,
    )
    def post(self, request, *args, **kwargs):
        process = self.get_object()
        process.retry()
        serializer = self.get_serializer(process)
        return Response(serializer.data)


@extend_schema(
    responses={201: DataImportSerializer},
    tags=['files'],
)
class DataImportFromFiles(CreateAPIView):
    """
    Start a data import from one or more uploaded files.
    """

    permission_classes = (IsVerified, )
    serializer_class = DataImportFromFilesSerializer

    def create(self, *args, **kwargs):
        super().create(*args, **kwargs)
        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(self.dataimport, context={'request': self.request}).data,
        )

    def perform_create(self, serializer):
        files = serializer.validated_data['files']
        corpus = files[0].corpus
        mode = serializer.validated_data['mode']
        # Serializer validation codes returns an element for 'folder_id'
        folder = serializer.validated_data.get('folder_id')
        folder_type = serializer.validated_data.get('folder_type')
        element_type = serializer.validated_data['element_type']

        if folder and folder.corpus != corpus:
            # The files' corpus is already validated as writable
            raise ValidationError({'__all__': ['Element and files are in different corpora']})

        # Start a dataimport with thumbnails generation
        self.dataimport = corpus.imports.create(
            creator=self.request.user,
            mode=mode,
            element=folder,
            folder_type=folder_type,
            element_type=element_type,
        )
        self.dataimport.files.set(files)
        self.dataimport.start(thumbnails=True)


@extend_schema_view(
    post=extend_schema(
        operation_id='CreateElementsWorkflow',
        tags=['imports'],
        responses={201: DataImportSerializer},
    )
)
class CorpusWorkflow(SelectionMixin, CorpusACLMixin, CreateAPIView):
    """
    Create a distributed workflow from elements of an Arkindex corpus.\n\n
    Requires an **admin** access to the corpus.
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementsWorkflowSerializer

    @transaction.atomic
    def create(self, request, pk=None, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        corpus, element, process_name, name_contains, element_type, selection, best_class, load_children, use_cache = map(
            lambda key: serializer.validated_data.pop(key, None),
            (
                'corpus',
                'element',
                'name',
                'element_name_contains',
                'element_type',
                'selection',
                'best_class',
                'load_children',
                'use_cache',
            )
        )

        process = corpus.imports.create(
            creator=self.request.user,
            mode=DataImportMode.Workers,
            name=process_name,
            name_contains=name_contains,
            element_type=element_type,
            element=element,
            best_class=best_class,
            load_children=load_children,
            use_cache=use_cache,
        )

        # Ensure process elements do exists
        if selection:
            elements = self.get_selection(corpus.id)
            if not elements.exists():
                raise ValidationError({
                    '__all__': ['No element match those filters.']
                })
            process.elements.set(elements)
        elif not process.list_elements().exists():
            raise ValidationError({
                '__all__': ['No element match those filters.']
            })

        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(
                process,
                context=self.get_serializer_context()
            ).data,
        )


class StartProcess(CorpusACLMixin, APIView):
    """
    Start a process, used to build a Workflow with Workers.\n\n
    Requires an **admin** access to the corpus of this process.
    """
    permission_classes = (IsVerified, )
    # For OpenAPI type discovery
    queryset = DataImport.objects.none()

    @extend_schema(
        operation_id='StartProcess',
        tags=['imports'],
        request=StartProcessSerializer,
        responses=DataImportSerializer
    )
    def post(self, request, pk=None, **kwargs):

        qs = DataImport.objects \
            .select_related('corpus') \
            .filter(corpus_id__isnull=False)

        process = get_object_or_404(qs, pk=self.kwargs['pk'])

        if not self.has_admin_access(process.corpus):
            raise PermissionDenied(detail='You do not have an admin access to the corpus of this process.')

        if process.mode != DataImportMode.Workers or process.workflow is not None:
            raise ValidationError(
                {'__all__': ['Only a DataImport with Workers mode and not already launched can be started later on']})

        serializer = StartProcessSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        process.start(**serializer.validated_data)

        return Response(
            status=status.HTTP_200_OK,
            data=DataImportSerializer(
                process,
                context={'request': request}
            ).data,
        )


@extend_schema(tags=['files'])
class DataFileList(CorpusACLMixin, ListAPIView):
    """
    List uploaded files in a corpus
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer
    # Tell the OpenAPI schema generator to believe this view is a list
    action = 'List'
    # For OpenAPI type discovery
    queryset = DataFile.objects.none()

    def get_queryset(self):
        return DataFile.objects.filter(corpus=self.get_corpus(self.kwargs['pk']))


@extend_schema(tags=['files'])
@extend_schema_view(
    get=extend_schema(description="Get an uploaded file's metadata"),
    patch=extend_schema(description="Update a datafile's status"),
    put=extend_schema(description="Update a datafile's status"),
    delete=extend_schema(description='Delete an uploaded file')
)
class DataFileRetrieve(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get one file
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer

    def get_queryset(self):
        # We need to use the default database to avoid stale read
        # when a datafile is updated immediately after it's created on a small upload
        return DataFile.objects.using('default').filter(
            corpus__in=Corpus.objects.readable(self.request.user)
        )

    def perform_update(self, serializer):
        if not self.has_write_access(serializer.instance.corpus):
            raise PermissionDenied
        return super().perform_update(serializer)

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        try:
            instance.s3_delete()
        except Exception as e:
            logger.warning(f'Could not delete DataFile from S3: {e}')
        return super().perform_destroy(instance)


@extend_schema_view(post=extend_schema(operation_id='CreateDataFile', tags=['files']))
class DataFileCreate(CreateAPIView):
    """
    Create a Datafile.
    In case of success, a signed URI is returned to upload file content directly to a remote server.
    """

    permission_classes = (IsVerified, )
    serializer_class = DataFileCreateSerializer


@extend_schema(exclude=True)
class GitRepositoryImportHook(APIView):
    """
    This endpoint is intended as a webhook for Git repository hosting applications like GitLab.
    """

    def post(self, request, pk=None, **kwargs):
        repo = get_object_or_404(Repository, id=pk)
        if not repo.enabled:
            raise PermissionDenied(detail='No credentials available for this repository.')

        try:
            repo.provider_class(credentials=repo.credentials).handle_webhook(repo, request)
        except Exception as e:
            user = repo.credentials.user
            # Notifying the user by mail that there was an error during webhook handling
            sent = send_mail(
                subject='An error occurred during Git hook handling',
                message=render_to_string(
                    'webhook_error.html',
                    context={
                        'user': user,
                        'url': repo.url,
                        'error': e.args[0],
                    },
                    request=request,
                ),
                from_email=None,
                recipient_list=[user.email],
                fail_silently=True,
            )
            if sent == 0:
                logger.error(f'Failed to send webhook error email to {user.email}')

            raise

        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema_view(
    get=extend_schema(
        tags=['repos'],
        operation_id='ListRepositories',
        description='List connected repositories',
    )
)
class RepositoryList(RepositoryACLMixin, ListAPIView):
    """
    List connected repositories
    """
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return self.readable_repositories \
            .annotate(authorized_users=Count('memberships')) \
            .select_related('credentials') \
            .prefetch_related('corpora', 'workers') \
            .order_by('url')


@extend_schema_view(
    get=extend_schema(
        operation_id='ListExternalRepositories',
        parameters=[
            OpenApiParameter(
                'search',
                description='Optional query terms to filter repositories',
            )
        ],
        tags=['repos'],
    ),
    post=extend_schema(
        operation_id='CreateExternalRepository',
        description='Using the given OAuth credentials, this links an external Git repository '
                    'to Arkindex, connects a push hook and starts an initial import.',
        responses={201: DataImportSerializer},
        tags=['repos'],
    )
)
class AvailableRepositoriesList(ListCreateAPIView):
    """
    List repositories associated to user OAuth credentials

    Using the given OAuth credentials ID, this uses the Git hosting
    application API's search feature to look for a repository matching
    the given query. Without a query, returns a full list.
    """
    permission_classes = (IsVerified, )
    pagination_class = None
    serializer_class = ExternalRepositorySerializer
    queryset = OAuthCredentials.objects.none()

    def get_queryset(self):
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])
        provider = cred.git_provider_class(credentials=cred)
        repos = provider.list_repos(query=self.request.query_params.get('search'))
        return map(provider.to_dict, repos)

    @transaction.atomic
    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = self.request.user
        cred = get_object_or_404(OAuthCredentials, user=user, id=self.kwargs['pk'])
        provider = cred.git_provider_class(credentials=cred)
        repo = provider.create_repo(
            request=self.request,
            id=serializer.validated_data.get('id')
        )
        # Add an admin membership to the repository creator
        repo.memberships.create(user=user, level=Role.Admin.value)

        latest_commit_sha = provider.get_latest_commit_sha(repo)
        rev, _ = provider.get_or_create_revision(repo, latest_commit_sha)
        extra_kwargs = {}

        # In case of a IIIF import, a corpus should be created and referenced to this repository
        if repo.type == RepositoryType.IIIF:
            corpus = Corpus.objects.create(
                name=f'IIIF import from {repo.url}',
                description=f'Automatic import from IIIF repository ({repo.url})',
                repository=repo
            )
            # Give the creator a write right on the corpus
            corpus.memberships.create(user=user, level=Role.Admin.value)
            # Add default types
            corpus.create_default_types()
            extra_kwargs['corpus_id'] = corpus.id

        dataimport = DataImport.objects.create(
            revision=rev,
            mode=DataImportMode.Repository,
            creator=cred.user,
            **extra_kwargs
        )
        # Generate thumbnails for IIIF repositories revisions only
        gen_thumbnails = repo.type == RepositoryType.IIIF
        dataimport.start(thumbnails=gen_thumbnails, **extra_kwargs)

        # Return the serialized data import
        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(dataimport, context={'request': self.request}).data,
        )


@extend_schema(
    tags=['repos'],
    description='Retrieve a repository',
)
@extend_schema_view(
    delete=extend_schema(description="Delete a repository")
)
class RepositoryRetrieve(RepositoryACLMixin, RetrieveDestroyAPIView):
    """
    Retrieve a repository
    """
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return self.readable_repositories.annotate(authorized_users=Count('memberships'))

    def check_object_permissions(self, request, repo):
        if self.request.method not in permissions.SAFE_METHODS and not self.has_admin_access(repo):
            raise PermissionDenied(detail='You do not have admin access to this repository.')


@extend_schema(
    tags=['repos'],
    description='Retrieve a revision',
)
class RevisionRetrieve(RepositoryACLMixin, RetrieveAPIView):
    """
    Retrieve a revision
    """
    permission_classes = (IsVerified, )
    serializer_class = RevisionSerializer

    def get_queryset(self):
        return Revision.objects.filter(repo__in=self.executable_repositories)


@extend_schema(tags=['repos'])
@extend_schema_view(
    get=extend_schema(operation_id='ListWorkersDeprecated'),
    post=extend_schema(operation_id='CreateWorkerDeprecated'),
)
class DeprecatedWorkerList(DeprecatedMixin, ListCreateAPIView):
    """
    List workers for a given repository UUID

    Create a Worker instance on a repository for given slug, name and type, return existing
    query if a worker with the same slug already exists on the repository
    """
    permission_classes = (IsVerified, )
    pagination_class = None
    deprecation_message = (
        'Listing workers associated to a repository is now deprecated. '
        'A user is able to list all the workers for which they have an execution access '
        'using `WorkerList` endpoint'
    )


@extend_schema(tags=['repos'])
@extend_schema_view(
    get=extend_schema(
        description='List workers for which the user has an execution access.',
        parameters=[
            OpenApiParameter(
                'repository_id',
                type=UUID,
                description='Filter workers by repository ID.',
                required=False,
            ),
            OpenApiParameter(
                'name',
                type=str,
                description='Filter workers by name.',
                required=False,
            )
        ]
    ),
    post=extend_schema(
        description=(
            'Create a worker.\n\n'
            'The user must be **internal** to perform this request.'
        )
    ),
)
class WorkerList(WorkerACLMixin, ListCreateAPIView):
    """
    List workers executable by the request user.
    Create a worker.
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerSerializer
    queryset = Worker.objects.none()

    def filter_queryset(self, qs):
        repo_id = self.request.query_params.get('repository_id')
        if repo_id:
            try:
                repo_id = UUID(repo_id)
            except (TypeError, ValueError):
                raise ValidationError({'repository_id': ['Invalid UUID']})
            qs = qs.filter(repository_id=repo_id)

        name_filter = self.request.query_params.get('name')
        if name_filter:
            qs = qs.filter(name__icontains=name_filter)

        return super().filter_queryset(qs)

    def get_queryset(self):
        """
        List executable workers (i.e. via a direct write right or a write right on its repository).
        """
        return self.executable_workers \
            .order_by('name', 'id') \
            .distinct()

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        repo_id = serializer.validated_data['repository_id']
        repo = get_object_or_404(Repository, pk=repo_id)

        access_right = get_max_level(self.request.user, repo)
        if not access_right:
            raise NotFound
        elif access_right < Role.Admin.value:
            raise PermissionDenied(detail='You do not have admin access to this repository.')

        worker, created = repo.workers.get_or_create(
            slug=serializer.validated_data['slug'],
            defaults={
                'name': serializer.validated_data['name'],
                'type': serializer.validated_data['type'],
            }
        )

        reponse_status = status.HTTP_201_CREATED if created else status.HTTP_200_OK

        return Response(WorkerSerializer(worker).data, status=reponse_status)


@extend_schema(tags=['repos'])
@extend_schema_view(
    get=extend_schema(
        description=(
            'List versions for a given worker ID with their revision and associated git references.\n\n'
            'Requires an **execution** access to the worker or its repository.'
        ),
        parameters=[
            OpenApiParameter(
                'mode',
                type=str,
                default='complete',
                description=('If set to `simple`, exclude versions older than master or with no tag')
            )
        ],
    ),
    post=extend_schema(
        description=(
            'Create a version for a given worker ID, revision ID and JSON configuration.\n\n'
            'The user must be **internal** to perform this request.'
        )
    ),
)
class WorkerVersionList(WorkerACLMixin, ListCreateAPIView):
    """
    List versions, their revision and associated git references for a given worker UUID

    Create a WorkerVersion instance for a worker and a revision with a given JSON configuration,
    return existing query if a workerVersion already exists for this worker and this revision
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerVersionSerializer
    queryset = WorkerVersion.objects.none()

    @property
    def simple_mode(self):
        return self.request.query_params.get('mode', 'complete').lower() == 'simple'

    def get_queryset(self):
        worker = get_object_or_404(
            Worker.objects.select_related('repository'),
            pk=self.kwargs['pk']
        )

        if not self.has_execution_access(worker):
            raise PermissionDenied(detail='You do not have an execution access to this worker.')

        filters = Q()
        if self.simple_mode:
            # Limit output to versions with tags or master branch
            filters = Q(revision__refs__type=GitRefType.Tag) | Q(revision__refs__type=GitRefType.Branch, revision__refs__name="master")

        return worker.versions \
            .filter(filters) \
            .prefetch_related('revision', 'revision__refs') \
            .distinct() \
            .order_by('-revision__created')

    def create(self, request, *args, **kwargs):
        if not self.request.user.is_internal:
            raise PermissionDenied(detail='Only internal users can create a new version from a worker.')

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        # We need to use the default database to avoid stale read
        # when creating a worker version on a newly created worker
        worker = get_object_or_404(
            Worker.objects.using('default'),
            pk=self.kwargs['pk']
        )

        revision = serializer.validated_data['revision']
        if worker.repository_id != revision.repo_id:
            raise ValidationError({
                'revision': ['The revision must be part of the same repository as the worker.']
            })

        version, created = WorkerVersion.objects.get_or_create(
            worker=worker,
            revision=revision,
            defaults={
                'configuration': serializer.validated_data['configuration'],
            }
        )

        reponse_status = status.HTTP_201_CREATED if created else status.HTTP_200_OK

        return Response(WorkerVersionSerializer(version).data, status=reponse_status)


@extend_schema_view(
    get=extend_schema(
        tags=['ml'],
        operation_id='ListCorpusWorkerVersions',
        description=(
            'List worker versions used by elements of a given corpus.\n\n'
            'No check is performed on workers access level in order to allow any user to see versions.'
        ),
    )
)
class CorpusWorkerVersionList(CorpusACLMixin, ListAPIView):
    """
    List worker versions used by elements of a given corpus.
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = WorkerVersionSerializer
    # For OpenAPI type discovery
    queryset = WorkerVersion.objects.none()

    def get_corpus(self):
        return get_object_or_404(self.readable_corpora, pk=self.kwargs['pk'])

    def get_queryset(self):
        corpus = self.get_corpus()
        return WorkerVersion.objects \
            .filter(elements__corpus_id=corpus.id) \
            .select_related('revision__repo', 'worker', 'worker__repository') \
            .prefetch_related('revision__refs', 'revision__versions') \
            .order_by('-revision__created') \
            .annotate(element_count=Count('id'))


@extend_schema(tags=['repos'])
class WorkerVersionRetrieve(RetrieveUpdateAPIView):
    """
    Retrieve a specific worker version. No authentication is required.
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = WorkerVersionEditSerializer
    queryset = WorkerVersion.objects.select_related('worker').all()

    def check_object_permissions(self, request, instance):
        """
        Allow any user to retrieve a worker version without authentication
        This behavior allows to retrieve a version information on public ressources

        Only internal users are allowed to update a worker version
        """
        if self.request.method in permissions.SAFE_METHODS:
            return
        super().check_object_permissions(request, instance)
        if not request.user.is_internal:
            raise PermissionDenied(detail='Only internal users are allowed to update a worker version.')


@extend_schema(
    tags=['repos'],
    description='Retrieve a worker',
)
class WorkerRetrieve(WorkerACLMixin, RetrieveAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerSerializer
    queryset = Worker.objects.all()

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset()).prefetch_related('repository')
        worker = get_object_or_404(queryset, pk=self.kwargs['pk'])
        if not self.has_read_access(worker):
            raise PermissionDenied(detail='You do not have a guest access to this worker.')
        super().check_object_permissions(self.request, worker)
        return worker


@extend_schema(tags=['imports'])
@extend_schema_view(
    get=extend_schema(description=(
        'List worker runs for a given process ID.\n\n'
        'Requires an **admin** access to the process project.'
    )),
    post=extend_schema(description=(
        'Create a worker run for a given process ID.\n\n'
        'Requires an **admin** access to the process project and a **contributor** access to the worker.'
    )),
)
class WorkerRunList(WorkerACLMixin, ListCreateAPIView):
    """
    List worker runs for a given process UUID.
    Create a run on a process.
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerRunSerializer
    queryset = WorkerRun.objects.none()

    def get_queryset(self):
        process = get_object_or_404(
            DataImport.objects.filter(corpus_id__isnull=False),
            pk=self.kwargs['pk']
        )

        if not self.has_access(process.corpus, Role.Admin.value):
            raise PermissionDenied(detail='You do not have an admin access to the corpus of this process.')

        return process.worker_runs \
            .select_related('version__worker') \
            .order_by('id')

    def perform_create(self, serializer):
        process = get_object_or_404(
            DataImport.objects.filter(corpus_id__isnull=False),
            pk=self.kwargs['pk']
        )

        # User must have admin access to the process project and a contributor access to the worker run
        if not self.has_access(process.corpus, Role.Admin.value):
            raise PermissionDenied(detail='You do not have an admin access to the corpus of this process.')
        try:
            worker = Worker.objects.get(versions__id=serializer.validated_data['version_id'])
        except Worker.DoesNotExist:
            raise ValidationError({'worker_version_id': ['This version does not exist.']})
        if not self.has_execution_access(worker):
            raise ValidationError({'worker_version_id': ['You do not have an execution access to this version.']})

        if process.mode != DataImportMode.Workers:
            raise ValidationError({'dataimport': ['Import mode must be Workers']})

        if process.workflow_id is not None:
            raise ValidationError({'__all__': ["Cannot create a WorkerRun on a DataImport that has already started"]})

        serializer.save(dataimport=process)


@extend_schema(tags=['imports'])
@extend_schema_view(
    get=extend_schema(
        operation_id='RetrieveWorkerRun',
        description='Retrieve details of a worker run.\n\nRequires an **admin** access to the process project.'
    ),
    patch=extend_schema(
        operation_id='PartialUpdateWorkerRun',
        description='Partially update a worker run.\n\nRequires an **admin** access to the process project.'
    ),
    put=extend_schema(
        operation_id='UpdateWorkerRun',
        description='Update dependencies of a worker run.\n\nRequires an **admin** access to the process project.'
    ),
    delete=extend_schema(
        operation_id='DestroyWorkerRun',
        description='Delete a worker run.\n\nRequires an **admin** access to the process project.'
    ),
)
class WorkerRunDetails(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve or edit a worker run i.e. update its dependencies.
    Requires an admin access to the process project.
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerRunEditSerializer

    def get_queryset(self):
        # Use default DB to avoid a race condition checking process workflow
        return WorkerRun.objects \
            .filter(dataimport__corpus_id__isnull=False) \
            .using('default') \
            .select_related('version__worker', 'dataimport__workflow', 'dataimport__corpus')

    def check_object_permissions(self, request, worker_run):
        if not self.has_admin_access(worker_run.dataimport.corpus):
            raise PermissionDenied(detail='You do not have an admin access to the process project.')

        # Updating a worker run is not possible once the process is started
        if request.method not in permissions.SAFE_METHODS and worker_run.dataimport.workflow_id is not None:
            raise ValidationError({'__all__': ["Cannot update a WorkerRun on a DataImport that has already started"]})

        super().check_object_permissions(request, worker_run)

    def perform_destroy(self, instance):
        # Remove references to this instance from its children parents field
        instance.dataimport.worker_runs.filter(parents__contains=[instance.id]).update(parents=ArrayRemove('parents', instance.id))
        return super().perform_destroy(instance)


@extend_schema_view(post=extend_schema(
    operation_id='CreateImportTranskribus',
    tags=['imports'],
    responses={
        201: DataImportSerializer,
        400: CreateImportTranskribusErrorResponseSerializer
    },
    examples=[OpenApiExample(
        status_codes=['400'],
        response_only=True,
        name="user-permission",
        value={'collection_id': "User user@example.com is not a member of the collection 1"},
        description="An error where the user is not a member of the collection."
    )]
))
class ImportTranskribus(CreateAPIView):
    """
    Start a data import from Transkribus email and collection ID.
    """

    permission_classes = (IsVerified, )
    serializer_class = ImportTranskribusSerializer

    def create(self, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES['transkribus']:
            raise ValidationError(['Transkribus import is unavailable due to the transkribus feature being disabled.'])
        super().create(*args, **kwargs)
        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(self.dataimport, context={'request': self.request}).data,
        )

    def perform_create(self, serializer):
        collection_id = serializer.validated_data['collection_id']

        # Create corpus
        corpus = Corpus.objects.create(
            name=f"Transkribus collection n°{collection_id}",
        )
        corpus.memberships.create(user=self.request.user, level=Role.Admin.value)
        corpus.create_default_transkribus_types()

        # Start a dataimport with thumbnails generation
        self.dataimport = corpus.imports.create(
            creator=self.request.user,
            mode=DataImportMode.Transkribus,
            collection_id=collection_id,
            build_entities=serializer.validated_data['build_entities']
        )
        self.dataimport.start(thumbnails=True)


@extend_schema_view(get=extend_schema(
    operation_id='ListProcessElements',
    parameters=[
        OpenApiParameter(
            'with_image',
            type=bool,
            default=False,
            description='Include zone and image information in the response.',
        )
    ],
    responses=PolymorphicProxySerializer(
        component_name='ListProcessElementsResponse',
        serializers={
            True: ProcessElementSerializer,
            False: ProcessElementLightSerializer,
        },
        resource_type_field_name='with_image',
    ),
    tags=['imports']
))
class ListProcessElements(CustomPaginationViewMixin, CorpusACLMixin, ListAPIView):
    """
    List all elements for a process with workers.\n\n
    Requires an **admin** access to the process corpus.
    """
    permission_classes = (IsVerified, )
    # For OpenAPI type discovery
    queryset = Element.objects.none()

    @property
    def with_image(self):
        return self.request.query_params.get('with_image', 'false').lower() in ('true', '1')

    @cached_property
    def process(self):
        return get_object_or_404(
            # Avoid stale read on newly created dataimports
            DataImport.objects.using('default').filter(corpus_id__isnull=False),
            Q(pk=self.kwargs['pk'])
        )

    def get_serializer_class(self):
        if self.with_image:
            return ProcessElementSerializer
        else:
            return ProcessElementLightSerializer

    def get_queryset(self):
        if not self.has_admin_access(self.process.corpus):
            raise PermissionDenied(detail='You do not have an admin access to the corpus of this process.')

        try:
            queryset = self.process.list_elements()
        except AssertionError as e:
            raise ValidationError({'__all__': [str(e)]})

        if not self.with_image:
            return queryset.values('id', 'type__slug', 'name')

        return queryset.annotate(
            # Build the image URL by concatenating the server's URL to the image's path
            # Server URLs might end in a slash, but not all the time,
            # so we strip any trailing slashes and add our own.
            # This uses NullIf as this concatenation will return '/' if there is no zone on the element.
            image_url=NullIf(
                Concat(
                    RTrimChr('zone__image__server__url', Value('/')),
                    Value('/'),
                    Cast('zone__image__path', CharField()),
                ),
                Value('/'),
            )
        ).values(
            'id',
            'type__slug',
            'name',
            'zone__image_id',
            'zone__image__width',
            'zone__image__height',
            'zone__polygon',
            'image_url',
        )


@extend_schema(tags=['ml'])
class UpdateWorkerActivity(GenericAPIView):
    """
    Makes a worker (internal user) able to update its activity on an element
    Only allow defined evolutions of the element's state
    """
    permission_classes = (IsInternalUser, )
    serializer_class = WorkerActivitySerializer

    @cached_property
    def allowed_transitions(self):
        # Defines a list of allowed previous states for any transition
        queued = WorkerActivityState.Queued.value
        started = WorkerActivityState.Started.value
        error = WorkerActivityState.Error.value
        processed = WorkerActivityState.Processed.value
        return {
            queued: [error],
            started: [queued, error],
            error: [started],
            processed: [started],
        }

    @extend_schema(
        operation_id='UpdateWorkerActivity',
        description=(
            'Updates the activity of a worker version on an element.\n\n'
            'The user must be **internal** to perform this request.'
        ),
    )
    def put(self, request, *args, **kwarg):
        serializer = self.get_serializer(data=request.data, partial=False)
        serializer.is_valid(raise_exception=True)

        worker_version_id = self.kwargs['pk']
        element_id = serializer.validated_data['element_id']
        state = serializer.validated_data['state'].value

        # We use the fact that only one worker activity may match the filter due to DB constraint
        activity = WorkerActivity.objects.filter(
            worker_version_id=worker_version_id,
            element_id=element_id,
            state__in=self.allowed_transitions[state]
        )
        update_count = activity.update(state=state)

        if not update_count:
            # As no row has been updated the provided data was incorrect
            raise ValidationError({
                '__all__': [
                    'Either this worker activity does not exists '
                    f"or updating the state to '{state}' is forbidden."
                ]
            })

        return Response(serializer.data)
