import logging
import uuid

from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.postgres.fields import HStoreField
from django.contrib.postgres.indexes import GinIndex
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.db.models import Q
from django.utils.functional import cached_property
from enumfields import Enum, EnumField

from arkindex.documents.dates import InterpretedDateMixin
from arkindex.documents.deletion import delete_element
from arkindex.documents.managers import CorpusManager, ElementManager
from arkindex.project.default_corpus import DEFAULT_CORPUS_TYPES, DEFAULT_TRANSKRIBUS_TYPES
from arkindex.project.elastic import ESElement, ESEntity, ESTranscription
from arkindex.project.fields import ArrayField
from arkindex.project.models import IndexableModel

logger = logging.getLogger(__name__)


class Corpus(IndexableModel):
    '''
    A group of elements, usually linked to users or projects
    '''
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    name = models.CharField(max_length=250)
    description = models.TextField(default="")
    repository = models.ForeignKey(
        'dataimport.Repository',
        on_delete=models.SET_NULL,
        related_name='corpora',
        null=True,
        blank=True
    )
    memberships = GenericRelation('users.Right', 'content_id')
    # Restrict top level elements to a certain type.
    # Efficient on large corpora as the type filter is applied first, assuming most elements of this type have no parent.
    top_level_type = models.ForeignKey(
        'documents.ElementType',
        on_delete=models.SET_NULL,
        related_name='top_level_corpora',
        null=True,
        blank=True
    )

    # Is this corpus publicly readable ?
    public = models.BooleanField(default=False)

    # Is this corpus indexable ?
    indexable = models.BooleanField(default=False)

    # Specific manager for ACL
    objects = CorpusManager()

    class Meta:
        verbose_name_plural = 'corpora'

    def __str__(self):
        return self.name

    def create_default_types(self):
        self.types.bulk_create(
            ElementType(corpus=self, **values)
            for values in DEFAULT_CORPUS_TYPES
        )

    def create_default_transkribus_types(self):
        self.types.bulk_create(
            ElementType(corpus=self, **values)
            for values in DEFAULT_TRANSKRIBUS_TYPES
        )


class ElementType(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    corpus = models.ForeignKey(Corpus, related_name='types', on_delete=models.CASCADE)
    slug = models.SlugField()
    display_name = models.CharField(max_length=250)
    folder = models.BooleanField(default=False)
    indexable = models.BooleanField(default=False)

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['corpus', 'slug'], name='corpus_unique_type_slug'),
        ]

    def __str__(self):
        return self.display_name


class ElementPath(models.Model):
    """
    Rolling cache table, with full paths exposed for each element
    An element can have multiple paths up to the top
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    element = models.ForeignKey('documents.Element', related_name='paths', on_delete=models.CASCADE)
    path = ArrayField(
        models.UUIDField(),
    )
    ordering = models.PositiveIntegerField(default=0)

    class Meta:
        indexes = [
            GinIndex(fields=['path']),
        ]


class Element(IndexableModel):
    """
    aka Unité Codicologique
    """
    corpus = models.ForeignKey('documents.Corpus', related_name='elements', on_delete=models.CASCADE)
    # `type` will replace `type` once all of Arkindex switches to generic element types
    type = models.ForeignKey(ElementType, related_name='elements', on_delete=models.PROTECT)
    name = models.CharField(max_length=250)
    zone = models.ForeignKey(
        'images.Zone',
        on_delete=models.SET_NULL,
        related_name='elements',
        null=True,
        blank=True,
    )
    worker_version = models.ForeignKey(
        'dataimport.WorkerVersion',
        on_delete=models.SET_NULL,
        related_name='elements',
        null=True,
        blank=True,
    )

    es_document = ESElement
    objects = ElementManager()

    def delete(self, *args, **kwargs):
        """
        Use raw SQL for element deletion and bypass Django's cascading
        """
        assert self.id, 'Cannot delete an element without an ID'
        delete_element(self.id)
        Element.objects.filter(id=self.id)._raw_delete(using='default')

    @transaction.atomic(using="default")
    def add_parent(self, parent, skip_children=False):
        '''
        Add an element as ancestor
        '''
        assert isinstance(parent, Element)

        # Load existing paths on current element, and the parent paths
        # in a single query
        paths = ElementPath.objects.filter(element_id__in=(self.id, parent.id)).values('element_id', 'path')
        existing_paths = [p['path'] for p in paths if p['element_id'] == self.id]
        parent_paths = [p['path'] for p in paths if p['element_id'] == parent.id]

        # Get the next order for this type in the parent
        order = parent.get_next_order(self.type)

        # Build all parent paths needed
        if not parent_paths:
            # No existing parent, adding new line
            parent_paths = [[]]

        new_paths = []
        for parent_path in parent_paths:

            # Build new full path and check it does not already exist
            path = parent_path + [parent.id]
            if path in existing_paths:
                continue

            # Check for cycle
            if self.id in path:
                raise ValueError('Cannot add a descendant as a parent')

            # Try to immediately insert
            new_paths.append(ElementPath(
                element=self,
                path=path,
                ordering=order,
            ))

        # Update children's paths in a non optimized way as this is pretty rare
        if skip_children is False:
            to_delete = []
            children = ElementPath.objects.filter(path__contains=[self.id])
            for child in children:
                base_path = child.path

                # This is a path with another parent, keep it
                if child.path[0] != self.id:
                    base_path = child.path[child.path.index(self.id):]  # Strip the other parent
                else:
                    to_delete.append(child.id)

                # Add new paths to create for the children
                for parent_path in parent_paths:
                    new_paths.append(ElementPath(
                        element=child.element,
                        path=parent_path + [parent.id] + base_path,
                        ordering=child.ordering,
                    ))

            # Clear old children paths, if any leftover
            if to_delete:
                ElementPath.objects.filter(id__in=to_delete).delete()

        # All new paths are finally created
        ElementPath.objects.bulk_create(new_paths)

    def get_next_order(self, type):
        """
        Find the next ordering for a new child in ElementPath
        Uses the primary database to avoid stale reads and duplicate orderings,
        and path__overlap to let Postgres use the GIN index before filtering by last item.
        """
        assert isinstance(type, ElementType)
        return ElementPath.objects \
                          .using('default') \
                          .filter(path__overlap=[self.id], path__last=self.id, element__type=type) \
                          .aggregate(max=models.Max('ordering') + 1)['max'] or 0

    @transaction.atomic
    def remove_child(self, child):
        '''
        Dissociate a child element
        '''
        assert isinstance(child, Element)

        # Check child is a direct child of this element
        if not ElementPath.objects.filter(element=child, path__last=self.id).exists():
            return

        new_epaths = {}

        # If child doesn't have any other parent, updates are needed
        if not ElementPath.objects.filter(element=child).exclude(path__last=self.id).exists():

            # Get all the child's children's paths
            children_epaths = list(ElementPath.objects.filter(
                element__in=list(Element.objects.get_descending(child.id)),
                path__contains=[self.id]
            ))

            for epath in children_epaths:
                new_path = epath.path[epath.path.index(self.id) + 1:]  # Strip this parent's path

                new_epaths[epath.element, tuple(new_path)] = ElementPath(
                    element=epath.element,
                    path=new_path,
                    ordering=epath.ordering,
                )

        # Delete everything
        ElementPath.objects.filter(
            element__in=list(Element.objects.get_descending(child.id)) + [child],
            path__contains=[self.id]
        ).delete()

        # Recreate paths as needed
        ElementPath.objects.bulk_create(new_epaths.values())

    @cached_property
    def thumbnail(self):
        from arkindex.images.models import Thumbnail  # Prevent circular imports
        return Thumbnail(self)

    def __str__(self):
        return '{}: {}'.format(self.type.display_name, self.name)


class EntityType(Enum):
    Person = 'person'
    Location = 'location'
    Subject = 'subject'
    Organization = 'organization'
    Misc = 'misc'
    Number = 'number'
    Date = 'date'


class Entity(InterpretedDateMixin, models.Model):
    """
    Semantic object in arkindex
    """
    es_document = ESEntity

    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    name = models.TextField()
    type = EnumField(EntityType, max_length=50, db_index=True)
    corpus = models.ForeignKey(Corpus, related_name='entities', on_delete=models.CASCADE)
    metas = HStoreField(null=True, blank=True)
    validated = models.BooleanField(default=False)
    moderator = models.ForeignKey(
        'users.User',
        related_name='entities',
        on_delete=models.SET_NULL,
        null=True,
        blank=True
    )
    worker_version = models.ForeignKey(
        'dataimport.WorkerVersion',
        on_delete=models.CASCADE,
        related_name='entities',
        null=True,
        blank=True,
    )

    class Meta:
        verbose_name_plural = 'Entities'

    def __str__(self):
        return self.name

    @property
    def raw_dates(self):
        if self.type != EntityType.Date:
            return
        return self.name


class EntityRole(models.Model):
    """
    Role's type between a parent and a child
    """
    parent_name = models.CharField(max_length=250)
    child_name = models.CharField(max_length=250)
    parent_type = EnumField(EntityType, max_length=50)
    child_type = EnumField(EntityType, max_length=50)
    corpus = models.ForeignKey(Corpus, related_name='roles', on_delete=models.CASCADE)

    class Meta:
        unique_together = (
            ('parent_name', 'child_name', 'parent_type', 'child_type', 'corpus'),
        )

    def __str__(self):
        return '{} -> {}'.format(self.parent_name, self.child_name)


class EntityLink(models.Model):
    """
    Link between two entities with a role
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    parent = models.ForeignKey(Entity, related_name='parents', on_delete=models.CASCADE)
    child = models.ForeignKey(Entity, related_name='children', on_delete=models.CASCADE)
    role = models.ForeignKey(EntityRole, related_name='links', on_delete=models.CASCADE)

    def clean(self):
        if self.role is None:
            return
        if self.parent is None:
            return
        if self.parent.type != self.role.parent_type:
            raise ValidationError("Parent's type {} is different from the expected type {}".format(
                self.parent.type,
                self.role.parent_type))
        if self.parent.corpus != self.role.corpus:
            raise ValidationError("Parent's corpus {} is different from the expected corpus {}".format(
                self.parent.corpus,
                self.role.corpus))
        if self.child is None:
            return
        if self.child.type != self.role.child_type:
            raise ValidationError("Child's type {} is different from the expected type {}".format(
                self.child.type,
                self.role.child_type))
        if self.child.corpus != self.role.corpus:
            raise ValidationError("Child's corpus {} is different from the expected corpus {}".format(
                self.child.corpus,
                self.role.corpus))

    def save(self, *args, **kwargs):
        self.full_clean()
        super().save(*args, **kwargs)


class Transcription(models.Model):
    """
    A transcription on:
     * a zone on an image
    """
    es_document = ESTranscription

    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    element = models.ForeignKey(
        Element,
        on_delete=models.CASCADE,
        related_name='transcriptions',
    )
    worker_version = models.ForeignKey(
        'dataimport.WorkerVersion',
        on_delete=models.CASCADE,
        related_name='transcriptions',
        null=True,
        blank=True,
    )
    text = models.TextField()
    confidence = models.FloatField(null=True, blank=True)
    entities = models.ManyToManyField(
        Entity,
        through='TranscriptionEntity',
        related_name='transcriptions',
    )

    def __str__(self):
        return 'Transcription: {}'.format(self.text[:20])


class TranscriptionEntity(models.Model):
    """
    Many-to-many relationship between Transcription and Entity
    """
    transcription = models.ForeignKey(
        Transcription,
        on_delete=models.CASCADE,
        related_name='transcription_entities',
    )
    entity = models.ForeignKey(
        Entity,
        on_delete=models.CASCADE,
        related_name='transcription_entities',
    )
    offset = models.PositiveIntegerField()
    length = models.PositiveIntegerField()

    class Meta:
        unique_together = (
            ('transcription', 'entity', 'offset', 'length'),
        )

    def clean(self):
        if self.transcription.element.corpus != self.entity.corpus:
            raise ValidationError('Transcription and Entity are not in the same corpus')
        if self.offset + self.length > len(self.transcription.text):
            raise ValidationError('Entity position overflows text size')
        super().clean()


class MLClass(models.Model):
    """
    A type of classification
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    name = models.CharField(max_length=1024)
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='ml_classes')

    class Meta:
        unique_together = (
            ('name', 'corpus'),
        )
        verbose_name_plural = 'classes'
        ordering = ('corpus', 'name')

    def __str__(self):
        return self.name


class ClassificationState(Enum):
    Pending = 'pending'
    Validated = 'validated'
    Rejected = 'rejected'


class Classification(models.Model):
    """
    A result of a classifier on a page
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    element = models.ForeignKey(
        Element,
        on_delete=models.CASCADE,
        related_name='classifications',
    )
    worker_version = models.ForeignKey(
        'dataimport.WorkerVersion',
        on_delete=models.CASCADE,
        related_name='classifications',
        null=True,
        blank=True,
    )
    moderator = models.ForeignKey(
        'users.User',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='classifications'
    )
    ml_class = models.ForeignKey(
        MLClass,
        on_delete=models.CASCADE,
        related_name='classifications'
    )
    # Predicted class is considered as correct by its creator
    high_confidence = models.BooleanField(default=False)
    state = EnumField(ClassificationState, max_length=16, default=ClassificationState.Pending, db_index=True)
    confidence = models.FloatField(null=True, blank=True)

    class Meta:
        constraints = [
            # Add class unicity for manual and non manual classifications on an element
            models.UniqueConstraint(
                fields=['element', 'ml_class'],
                name='classification_unique_manual',
                condition=Q(worker_version_id__isnull=True),
            ),
            models.UniqueConstraint(
                fields=['element', 'ml_class', 'worker_version'],
                name='classification_unique_worker_version',
                condition=Q(worker_version_id__isnull=False),
            )
        ]


class MetaType(Enum):
    Text = 'text'
    Markdown = 'markdown'
    Date = 'date'
    Location = 'location'
    # Element's original structure reference (intended to be indexed)
    Reference = 'reference'


class AllowedMetaData(models.Model):
    """
    Defines standard metadata's (type, name) couples for a corpus
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    corpus = models.ForeignKey('documents.Corpus', related_name='allowed_metadatas', on_delete=models.CASCADE)
    type = EnumField(MetaType, max_length=50)
    name = models.CharField(max_length=250)

    class Meta:
        ordering = ('corpus', 'type', 'name')
        unique_together = (
            ('corpus', 'type', 'name'),
        )


class MetaData(InterpretedDateMixin, models.Model):
    """
    Metadatas for elements
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    element = models.ForeignKey(Element, related_name='metadatas', on_delete=models.CASCADE)
    name = models.CharField(max_length=250)
    type = EnumField(MetaType, max_length=50, db_index=True)
    value = models.TextField()
    index = models.PositiveIntegerField(default=0)
    entity = models.ForeignKey(Entity, null=True, blank=True, related_name='metadatas', on_delete=models.SET_NULL)
    worker_version = models.ForeignKey(
        'dataimport.WorkerVersion',
        on_delete=models.CASCADE,
        related_name='metadatas',
        null=True,
        blank=True,
    )

    class Meta:
        ordering = ('element', 'name', 'index')
        unique_together = (
            ('element', 'name', 'index'),
        )

    @property
    def raw_dates(self):
        if self.type != MetaType.Date:
            return
        return self.value

    def clean(self):
        if self.entity is None or self.element is None:
            return
        if self.entity.corpus != self.element.corpus:
            raise ValidationError("Entity's corpus {} is different from the expected corpus {}".format(
                self.entity.corpus,
                self.element.corpus))

    def save(self, *args, **kwargs):
        self.clean()
        super().save(*args, **kwargs)


class Selection(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    element = models.ForeignKey(Element, related_name='selections', on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='selections', on_delete=models.CASCADE)

    class Meta:
        unique_together = (
            ('element', 'user'),
        )
