from rest_framework import serializers

from arkindex.documents.models import Corpus, Element
from arkindex.project.triggers import reindex_start


class ReindexConfigSerializer(serializers.Serializer):
    corpus = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.all(), required=False)
    element = serializers.PrimaryKeyRelatedField(queryset=Element.objects.all(), required=False)
    transcriptions = serializers.BooleanField(default=True)
    elements = serializers.BooleanField(default=True)
    entities = serializers.BooleanField(default=True)
    drop = serializers.BooleanField(default=False)

    default_error_messages = {
        'reindex_nothing': 'At least one index type is required.',
        'filtered_drop': '`drop` can only be used when reindexing everything.',
        'element_in_corpus': 'The selected element is not in the selected corpus.',
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            return
        user = self.context.get('request').user
        self.fields['corpus'].queryset = Corpus.objects.admin(user)
        self.fields['element'].queryset = Element.objects.filter(corpus__in=Corpus.objects.admin(user))

    def validate(self, data):
        corpus, element, transcriptions, elements, entities, drop = map(
            data.get,
            ('corpus', 'element', 'transcriptions', 'elements', 'entities', 'drop')
        )
        if not (transcriptions or elements or entities):
            self.fail('reindex_nothing')
        if drop and (corpus or element):
            self.fail('filtered_drop')
        if element and corpus and element.corpus_id != corpus.id:
            self.fail('element_in_corpus')
        return data

    def save(self):
        reindex_start(**self.validated_data, user_id=self.context['request'].user.id)
