from unittest.mock import patch

from django.core.management import CommandError, call_command
from django.test import override_settings

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus, EntityType, MetaType
from arkindex.project.tests import FixtureTestCase


@override_settings(ARKINDEX_FEATURES={'search_v2': True})
class TestReindexV2Command(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name='private', indexable=True)
        cls.worker_version = WorkerVersion.objects.first()
        cls.worker = cls.worker_version.worker

        # Create element types
        folder_type, _ = cls.private_corpus.types.get_or_create(slug='folder', display_name='Folder', folder=True)
        cls.page_type, _ = cls.private_corpus.types.get_or_create(slug='page', display_name='Page', indexable=True)
        cls.line_type, _ = cls.private_corpus.types.get_or_create(slug='text_line', display_name='Line')
        cls.word_type, _ = cls.private_corpus.types.get_or_create(slug='word', display_name='Word')

        # Create elements
        vol = cls.private_corpus.elements.create(name='Folder', type=folder_type)
        cls.page = cls.private_corpus.elements.create(name='New page', type=cls.page_type)
        cls.line = cls.private_corpus.elements.create(name='A line', type=cls.line_type)
        cls.page.add_parent(vol)
        cls.line.add_parent(cls.page)

    @patch("arkindex.documents.indexer_v2.solr")
    def test_run_empty_element(self, mock_solr):
        """
        Test the reindex command with no indexable type
        """
        self.page_type.indexable = False
        self.page_type.save()

        call_command('reindex_v2', self.private_corpus.id)
        self.assertEqual(mock_solr.delete_doc_by_query.call_count, 0)
        self.assertEqual(mock_solr.index.call_count, 0)

    @patch('arkindex.documents.indexer_v2.solr')
    def test_run_multiple_elements(self, mock_solr):
        """
        Test the reindex command for multiple elements
        """
        word = self.private_corpus.elements.create(name='A word', type=self.word_type)
        word.add_parent(self.line)

        call_command('reindex_v2', self.private_corpus.id)
        self.assertEqual(mock_solr.delete_doc_by_query.call_count, 0)
        self.assertEqual(mock_solr.index.call_count, 1)
        (index_name, documents), kwargs = mock_solr.index.call_args
        self.assertEqual(index_name, f'project-{self.private_corpus.id}')
        self.assertListEqual(documents, [
            {
                'element_id': str(self.page.id),
                'element_text': self.page.name,
                'element_type': self.page.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(word.id),
                'element_text': word.name,
                'element_type': word.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            }]
        )
        self.assertDictEqual(kwargs, {'commit': True})

    @patch('arkindex.documents.indexer_v2.solr')
    def test_run_transcriptions(self, mock_solr):
        """
        Test the reindex command for element with transcriptions
        """
        tr_1 = self.line.transcriptions.create(
            confidence=0.8,
            text='Transcription for the line',
            worker_version=self.worker_version,
        )
        tr_2 = self.line.transcriptions.create(
            confidence=0.5,
            text='Second transcription',
            worker_version=self.worker_version,
        )

        call_command('reindex_v2', self.private_corpus.id)
        self.assertEqual(mock_solr.delete_doc_by_query.call_count, 0)
        self.assertEqual(mock_solr.index.call_count, 1)
        (index_name, documents), kwargs = mock_solr.index.call_args
        self.assertEqual(index_name, f'project-{self.private_corpus.id}')
        self.assertListEqual(documents, [
            {
                'element_id': str(self.page.id),
                'element_text': self.page.name,
                'element_type': self.page.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'transcription_confidence': tr_1.confidence,
                'transcription_text': tr_1.text,
                'transcription_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}'
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'transcription_confidence': tr_2.confidence,
                'transcription_text': tr_2.text,
                'transcription_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}'
            }]
        )
        self.assertDictEqual(kwargs, {'commit': True})

    @patch('arkindex.documents.indexer_v2.solr')
    def test_run_classifications(self, mock_solr):
        """
        Test the reindex command for element with classifications
        """
        cl_1 = self.line.classifications.create(
            confidence=0.8,
            worker_version=self.worker_version,
            ml_class=self.private_corpus.ml_classes.create(name='Cat')
        )
        cl_2 = self.line.classifications.create(
            confidence=0.4,
            worker_version=self.worker_version,
            ml_class=self.private_corpus.ml_classes.create(name='Dog')
        )

        call_command('reindex_v2', self.private_corpus.id)
        self.assertEqual(mock_solr.delete_doc_by_query.call_count, 0)
        self.assertEqual(mock_solr.index.call_count, 1)
        (index_name, documents), kwargs = mock_solr.index.call_args
        self.assertEqual(index_name, f'project-{self.private_corpus.id}')
        self.assertListEqual(documents, [
            {
                'element_id': str(self.page.id),
                'element_text': self.page.name,
                'element_type': self.page.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'classification_name': cl_1.ml_class.name,
                'classification_confidence': cl_1.confidence,
                'classification_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'classification_name': cl_2.ml_class.name,
                'classification_confidence': cl_2.confidence,
                'classification_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            }
        ])
        self.assertDictEqual(kwargs, {'commit': True})

    @patch('arkindex.documents.indexer_v2.solr')
    def test_run_metadatas(self, mock_solr):
        """
        Test the reindex command for element with metadatas
        """
        self.private_corpus.allowed_metadatas.create(type=MetaType.Location, name='Country')
        self.private_corpus.allowed_metadatas.create(type=MetaType.Text, name='Folio')
        md_1 = self.line.metadatas.create(
            type=MetaType.Location,
            name='Country',
            value='France',
            worker_version=self.worker_version,
        )
        md_2 = self.line.metadatas.create(
            type=MetaType.Text,
            name='Folio',
            value='1',
            worker_version=self.worker_version,
        )

        call_command('reindex_v2', self.private_corpus.id)
        self.assertEqual(mock_solr.delete_doc_by_query.call_count, 0)
        self.assertEqual(mock_solr.index.call_count, 1)
        (index_name, documents), kwargs = mock_solr.index.call_args
        self.assertEqual(index_name, f'project-{self.private_corpus.id}')
        self.assertListEqual(documents, [
            {
                'element_id': str(self.page.id),
                'element_text': self.page.name,
                'element_type': self.page.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'metadata_text': md_1.value,
                'metadata_type': md_1.type.value,
                'metadata_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'metadata_text': md_2.value,
                'metadata_type': md_2.type.value,
                'metadata_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            }
        ])
        self.assertDictEqual(kwargs, {'commit': True})

    @patch('arkindex.documents.indexer_v2.solr')
    def test_run_entities(self, mock_solr):
        """
        Test the reindex command for element with entities
        """
        self.private_corpus.allowed_metadatas.create(type=MetaType.Location, name='Country')
        entity_1 = self.private_corpus.entities.create(name="CDLK", type=EntityType.Location, worker_version=self.worker_version)
        tr = self.line.transcriptions.create(
            confidence=0.8,
            text='Transcription for the line',
        )
        entity_1.transcription_entities.create(
            transcription=tr,
            offset=0,
            length=len(entity_1.name)
        )
        entity_2 = self.private_corpus.entities.create(name="Robert", type=EntityType.Person, worker_version=self.worker_version)
        md = self.line.metadatas.create(
            type=MetaType.Location,
            name='Country',
            value='France',
            entity=entity_2
        )

        call_command('reindex_v2', self.private_corpus.id)
        self.assertEqual(mock_solr.delete_doc_by_query.call_count, 0)
        self.assertEqual(mock_solr.index.call_count, 1)
        (index_name, documents), kwargs = mock_solr.index.call_args
        self.assertEqual(index_name, f'project-{self.private_corpus.id}')
        self.assertListEqual(documents, [
            {
                'element_id': str(self.page.id),
                'element_text': self.page.name,
                'element_type': self.page.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'transcription_confidence': tr.confidence,
                'transcription_text': tr.text,
                'transcription_worker': None
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'metadata_text': md.value,
                'metadata_type': md.type.value,
                'metadata_worker': None
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'entity_text': entity_1.name,
                'entity_type': entity_1.type.value,
                'entity_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
                'entity_text': entity_2.name,
                'entity_type': entity_2.type.value,
                'entity_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
            }
        ])
        self.assertDictEqual(kwargs, {'commit': True})

    @patch('arkindex.documents.indexer_v2.solr')
    def test_run_element_worker(self, mock_solr):
        """
        Test the reindex command for element with a worker version
        """
        self.line.worker_version = self.worker_version
        self.line.save()

        call_command('reindex_v2', self.private_corpus.id)
        self.assertEqual(mock_solr.delete_doc_by_query.call_count, 0)
        self.assertEqual(mock_solr.index.call_count, 1)
        (index_name, documents), kwargs = mock_solr.index.call_args
        self.assertEqual(index_name, f'project-{self.private_corpus.id}')
        self.assertListEqual(documents, [
            {
                'element_id': str(self.page.id),
                'element_text': self.page.name,
                'element_type': self.page.type.display_name,
                'element_worker': None,
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            },
            {
                'element_id': str(self.line.id),
                'element_text': self.line.name,
                'element_type': self.line.type.display_name,
                'element_worker': f'{self.worker.id.hex[:10]}-{self.worker.name}',
                'element_image': None,
                'parent_id': str(self.page.id),
                'parent_name': self.page.name,
                'parent_type': self.page.type.display_name,
            }
        ])
        self.assertDictEqual(kwargs, {'commit': True})

    @patch('arkindex.documents.indexer_v2.solr')
    def test_drop(self, mock_solr):
        """
        Test the reindex command can drop indexes
        """
        self.page_type.indexable = False
        self.page_type.save()

        call_command('reindex_v2', self.private_corpus.id, drop=True)
        self.assertEqual(mock_solr.delete_doc_by_query.call_count, 1)
        (index_name, query), kwargs = mock_solr.delete_doc_by_query.call_args
        self.assertEqual(index_name, f'project-{self.private_corpus.id}')
        self.assertEqual(query, '*:*')
        self.assertDictEqual(kwargs, {'commit': True})

    def test_corpus_not_found(self):
        corpus_id = 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'
        with self.assertRaises(CommandError) as context:
            call_command('reindex_v2', corpus_id)
        self.assertEqual(
            str(context.exception),
            f'Corpus {corpus_id} does not exist'
        )

    def test_corpus_not_indexable(self):
        self.private_corpus.indexable = False
        self.private_corpus.save()

        with self.assertRaises(CommandError) as context:
            call_command('reindex_v2', self.private_corpus.id)

        self.assertEqual(
            str(context.exception),
            f'Corpus {self.private_corpus.name} is not indexable'
        )

    @override_settings(ARKINDEX_FEATURES={'search_v2': False})
    def test_no_search(self):
        with self.assertRaises(CommandError) as context:
            call_command('reindex_v2', self.private_corpus.id)
        self.assertEqual(
            str(context.exception),
            'Reindexation is not possible if the search feature flag is disabled. '
            'Consider setting `features.search_v2` to `on` or `true` or `yes` in the YAML '
            'configuration file, and configuring Sorl properly.'
        )
