from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase


class TestBulkTranscriptions(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    def test_bulk_transcriptions_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:transcription-bulk'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_bulk_transcriptions_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:transcription-bulk'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_bulk_transcriptions_not_found(self):
        self.client.force_login(self.user)
        self.user.rights.all().delete()
        forbidden_element = self.corpus.elements.get(name='Volume 1, page 1r')
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:transcription-bulk'), {
                "worker_version": str(self.worker_version.id),
                "transcriptions": [
                    {
                        "element_id": str(forbidden_element.id),
                        "text": "lol",
                        "confidence": 0.4
                    },
                    {
                        "element_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                        "text": "lol",
                        "confidence": 0.4
                    }
                ],
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'transcriptions': [
                {'element_id': [f'Element {forbidden_element.id} was not found or cannot be written to.']},
                {'element_id': ['Element aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa was not found or cannot be written to.']},
            ]
        })

    def test_bulk_transcriptions(self):
        self.client.force_login(self.user)

        element1 = self.corpus.elements.get(name='Volume 2')
        element2 = self.corpus.elements.get(name='Volume 2, page 1r')
        self.assertFalse(element1.transcriptions.exists())
        self.assertFalse(element2.transcriptions.exists())

        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:transcription-bulk'), {
                "worker_version": str(self.worker_version.id),
                "transcriptions": [
                    {
                        "element_id": str(element1.id),
                        "text": "Sneasel",
                        "confidence": 0.54
                    },
                    {
                        "element_id": str(element2.id),
                        "text": "Charizard",
                        "confidence": 0.85
                    },
                    {
                        "element_id": str(element1.id),
                        "text": "Raticate",
                        "score": 0.12
                    },
                ],
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        first_tr, second_tr = element1.transcriptions.values_list('id', flat=True)
        third_tr = element2.transcriptions.values_list('id', flat=True)[0]
        self.assertEqual(response.json(), {
            "worker_version": str(self.worker_version.id),
            "transcriptions": [
                {
                    "id": str(first_tr),
                    "element_id": str(element1.id),
                    "text": "Sneasel",
                    "confidence": 0.54,
                },
                {
                    "id": str(third_tr),
                    "element_id": str(element2.id),
                    "text": "Charizard",
                    "confidence": 0.85,
                },
                {
                    "id": str(second_tr),
                    "element_id": str(element1.id),
                    "text": "Raticate",
                    "confidence": 0.12,
                },
            ]
        })

        self.assertCountEqual(
            list(element1.transcriptions.values('text', 'confidence')),
            [
                {
                    "text": "Sneasel",
                    "confidence": 0.54
                },
                {
                    "text": "Raticate",
                    "confidence": 0.12
                },
            ]
        )
        self.assertCountEqual(
            list(element2.transcriptions.values('text', 'confidence')),
            [{
                "text": "Charizard",
                "confidence": 0.85
            }]
        )

    def test_bulk_transcriptions_score_xor_confidence(self):
        self.client.force_login(self.user)
        element = self.corpus.elements.first()

        response = self.client.post(reverse('api:transcription-bulk'), {
            "worker_version": str(self.worker_version.id),
            "transcriptions": [
                {
                    "element_id": str(element.id),
                    "text": "Sneasel",
                    "score": 0.54,
                    "confidence": 0.54
                },
            ],
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'transcriptions': [
            {'non_field_errors': [
                'Either the score or confidence field should be defined.'
            ]}
        ]})

    def test_bulk_transcriptions_requires_confidence(self):
        self.client.force_login(self.user)
        element = self.corpus.elements.first()

        response = self.client.post(reverse('api:transcription-bulk'), {
            "worker_version": str(self.worker_version.id),
            "transcriptions": [
                {
                    "element_id": str(element.id),
                    "text": "Sneasel"
                },
            ],
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'transcriptions': [
            {'non_field_errors': [
                'Either the score or confidence field should be defined.'
            ]}
        ]})
