from django.core.management.base import BaseCommand
from django.db import connections, transaction
from teklia_toolbox.time import Timer


class Command(BaseCommand):
    """
    With the migration of zone polygons to PostGIS LineStrings, the new application of `ST_Normalize` did a better
    job at reordering than `Polygon.reorder`, causing duplicated zones in some databases. This command is intended
    to be run only if `images.0005` fails to create the unique constraint:

    arkindex migrate images 0004
    arkindex deduplicate_zones
    arkindex migrate images

    This is a modified version of the zone deduplication that was performed back when the polygon_hash column
    was being created: https://gitlab.com/arkindex/backend/-/snippets/1906353
    """

    help = 'Deduplicate zones using SQL. Requires the images.0004 migration'

    @transaction.atomic
    def handle(self, **options):
        with connections['default'].cursor() as cursor:
            self.stdout.write('Looking for duplicated zones…')
            with Timer() as t:
                cursor.execute("""
                CREATE TABLE duplicated_zones (id, unique_id)
                AS SELECT x.* FROM (
                    SELECT
                        id,
                        first_value(id) OVER (
                            PARTITION BY image_id, polygon
                            ORDER BY id ASC
                        ) AS unique_id
                    FROM images_zone
                ) AS x WHERE x.id != x.unique_id;
                """)

                cursor.execute('SELECT COUNT(*) FROM duplicated_zones;')
                count = cursor.fetchone()[0]

            self.stdout.write(f'{count} duplicated zones found in {t.delta}.')

            if count > 0:
                self.stdout.write('Unlinking duplicated zones from elements…')
                with Timer() as t:
                    cursor.execute("""
                    UPDATE documents_element
                    SET zone_id = duplicated_zones.unique_id
                    FROM duplicated_zones
                    WHERE duplicated_zones.id = documents_element.zone_id;
                    """)
                self.stdout.write(f'Duplicated zones unlinked from elements in {t.delta}.')

                self.stdout.write('Unlinking duplicated zones from transcriptions…')
                with Timer() as t:
                    cursor.execute("""
                    UPDATE documents_transcription
                    SET zone_id = duplicated_zones.unique_id
                    FROM duplicated_zones
                    WHERE duplicated_zones.id = documents_transcription.zone_id;
                    """)
                self.stdout.write(f'Duplicated zones unlinked from transcriptions in {t.delta}.')

                self.stdout.write('Removing duplicated zones…')
                with Timer() as t:
                    cursor.execute("""
                    DELETE FROM images_zone
                    USING duplicated_zones
                    WHERE images_zone.id = duplicated_zones.id;
                    """)
                self.stdout.write(f'Duplicated zones removed in {t.delta}.')

            self.stdout.write('Removing temporary table…')
            with Timer() as t:
                cursor.execute("DROP TABLE duplicated_zones;")
            self.stdout.write(f'Temporary table removed in {t.delta}.')

            if count:
                self.stdout.write(self.style.SUCCESS(f'{count} duplicated zones have been removed.'))

            self.stdout.write('Committing transaction…')
