import importlib
from unittest import TestCase
from unittest.mock import MagicMock, patch

from django.test import override_settings

from arkindex.project import aws

AWS_SETTINGS = {
    'AWS_ACCESS_KEY': 'golden_key',
    'AWS_SECRET_KEY': 'golden_ticket',
    'AWS_ENDPOINT': 'https://backdoor',
    'AWS_REGION': 'middle-earth',
}


class TestAWSMixin(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @override_settings(**AWS_SETTINGS)
    def test_session(self):
        importlib.reload(aws)
        self.assertEqual(aws.s3.meta.client._endpoint.host, 'https://backdoor')

        client_config = aws.s3.meta.client._client_config
        self.assertEqual(client_config.region_name, 'middle-earth')
        self.assertEqual(client_config.signature_version, 's3v4')

    @patch('arkindex.project.aws.logger')
    def test_null_bucket(self, logger_mock):
        class fileClass(aws.S3FileMixin):
            s3_bucket = None
        s3_file = fileClass()
        self.assertEqual(s3_file.get_s3_object(), None)
        self.assertEqual(s3_file.s3_put_url, None)
        self.assertEqual(s3_file.s3_url, None)
        self.assertEqual(logger_mock.info.call_count, 2)

    @patch('arkindex.project.aws.S3FileMixin.s3_object')
    def test_check_hash(self, s3_object_mock):
        s3_object_mock.e_tag = '"deadbeef"'
        save_mock = MagicMock()

        class fileClass(aws.S3FileMixin):
            s3_bucket = None
            hash = 'deadbeef'
            status = aws.S3FileStatus.Unchecked
            save = save_mock
        s3_file = fileClass()
        s3_file.check_hash(save=True)
        self.assertEqual(s3_file.status, aws.S3FileStatus.Checked)
        self.assertEqual(save_mock.call_count, 1)

    @patch('arkindex.project.aws.S3FileMixin.s3_object')
    def test_wrong_hash_raises(self, s3_object_mock):
        s3_object_mock.e_tag = '"dead"'
        save_mock = MagicMock()

        class fileClass(aws.S3FileMixin):
            s3_bucket = None
            hash = 'beef'
            status = aws.S3FileStatus.Unchecked
            save = save_mock
        s3_file = fileClass()
        with self.assertRaisesRegex(ValueError, 'MD5 hashes do not match'):
            s3_file.check_hash(save=True, raise_exc=True)
        self.assertEqual(s3_file.status, aws.S3FileStatus.Error)
        self.assertEqual(save_mock.call_count, 1)

    @patch('arkindex.project.aws.logger')
    @patch('arkindex.project.aws.S3FileMixin.s3_object')
    def test_check_hash_ignores_multipart(self, s3_object_mock, logger_mock):
        s3_object_mock.e_tag = '"badbadbadbad-5"'
        save_mock = MagicMock()

        class FileClass(aws.S3FileMixin):
            s3_bucket = None
            hash = 'beef'
            status = aws.S3FileStatus.Unchecked
            save = save_mock

        s3_file = FileClass()
        s3_file.check_hash(save=True, raise_exc=True)
        self.assertEqual(s3_file.status, aws.S3FileStatus.Unchecked)
        self.assertEqual(save_mock.call_count, 1)
        self.assertEqual(logger_mock.warning.call_count, 1)
