from django.conf import settings
from django.shortcuts import get_object_or_404
from django.views.generic import TemplateView, View
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import permissions
from rest_framework.exceptions import NotFound, PermissionDenied

from arkindex.dataimport.models import DataImport
from arkindex.dataimport.permissions import IsArtifactAdminOrCreator, IsTaskAdminOrCreator
from arkindex.project.mixins import CachedViewMixin, ProcessACLMixin
from arkindex.project.permissions import IsVerified
from arkindex.users.models import Role
from ponos.api import AgentDetails, AgentsState, SecretDetails, TaskArtifactDownload, TaskUpdate, WorkflowDetails


class FrontendView(View):
    """
    View that show the frontend's router
    TODO: Get rid of this view once the frontend is fully split
    """


class CdnHome(CachedViewMixin, TemplateView):
    """
    Index template to use frontend assets from a CDN
    """
    template_name = 'index.html'
    cache_timeout = 86400
    cache_prefix = settings.FRONTEND_VERSION

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx['cdn_assets_url'] = settings.CDN_ASSETS_URL
        ctx['mirador_url'] = settings.MIRADOR_URL
        ctx['universal_viewer_url'] = settings.UNIVERSAL_VIEWER_URL
        ctx['version'] = settings.VERSION
        ctx['frontend_version'] = settings.FRONTEND_VERSION
        ctx['frontend_sentry_dsn'] = settings.FRONTEND_SENTRY_DSN
        ctx['frontend_doorbell'] = settings.ARKINDEX_FEATURES['doorbell'] and settings.FRONTEND_DOORBELL
        return ctx


class OpenAPIDocsView(CachedViewMixin, TemplateView):
    """
    OpenAPI documentation via ReDoc
    """
    template_name = 'redoc.html'
    cache_timeout = 86400
    cache_prefix = settings.VERSION


class PonosSecretDetails(SecretDetails):
    """
    Allow Arkindex internal users to retrieve a secret as if they were Ponos agents
    """
    class IsInternalOnly(permissions.BasePermission):

        def has_permission(self, request, view):
            return request.user.is_authenticated and request.user.is_internal

    permission_classes = (IsInternalOnly, )


class PonosAgentsState(AgentsState):
    """
    Allow any verified user to see the state of Ponos agents on this instance
    """
    permission_classes = (IsVerified, )


class PonosAgentDetails(AgentDetails):
    """
    Allow any verified user to see the details of an agent including all its running tasks
    """
    permission_classes = (IsVerified, )


class PonosTaskUpdate(TaskUpdate):
    """
    Allow users with an admin privilege and creators to update a dataimport task
    """
    permission_classes = (IsTaskAdminOrCreator, )


class PonosArtifactDownload(TaskArtifactDownload):
    """
    Allow users with an admin privilege and creators to download an artifact of a dataimport task
    """
    permission_classes = (IsArtifactAdminOrCreator, )


@extend_schema_view(
    put=extend_schema(description="Update a workflow's status and tasks.\n\nRequires an **admin** right to its process."),
    patch=extend_schema(description='Partially update a workflow.\n\nRequires an **admin** right to its process.')
)
class PonosWorkflowDetails(ProcessACLMixin, WorkflowDetails):
    """
    Allow users with an admin access to a process to retry its workflow
    """
    permission_classes = (IsVerified, )

    def check_object_permissions(self, request, workflow):
        super().check_object_permissions(request, workflow)

        required_access = Role.Admin.value
        # Allow a user with a read access on the process to retrieve its workflow
        if request.method in permissions.SAFE_METHODS:
            required_access = Role.Guest.value

        process = get_object_or_404(
            DataImport,
            workflow_id=workflow.id
        )
        access_level = self.process_access_level(process)
        if not access_level:
            raise NotFound
        if access_level < required_access:
            raise PermissionDenied(detail='You do not have an admin access to this process.')
